/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.ui.view.fx;

import static javafx.scene.paint.Color.BLUE;
import static javafx.scene.paint.Color.DARKGRAY;
import static javafx.scene.paint.Color.GREEN;
import static javafx.scene.paint.Color.LIGHTGRAY;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_10PT;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_12PT;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_14PT;
import static org.fortiss.tooling.common.ui.javafx.style.FontStyle.BLACK_VERDANA_8PT;
import static org.fortiss.tooling.common.ui.javafx.style.LineStyle.SOLID_BLACK_1PT;
import static org.fortiss.tooling.kernel.ui.util.SelectionUtils.checkAndPickFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstParentWithType;

import java.text.DecimalFormat;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.common.ui.javafx.layout.CompositeFXControllerBase;
import org.fortiss.tooling.common.ui.javafx.style.FillStyle;
import org.fortiss.tooling.common.ui.javafx.style.FontStyle;
import org.fortiss.tooling.common.ui.javafx.style.LineStyle;
import org.fortiss.tooling.ext.quality.data.MetricTreeNode;
import org.fortiss.tooling.ext.quality.service.ModelQualityService;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.spiderchart.control.SpiderChartViewer;
import org.fortiss.tooling.spiderchart.model.DataSeries;
import org.fortiss.tooling.spiderchart.model.DoubleAxis;
import org.fortiss.tooling.spiderchart.model.SpiderChart;
import org.fortiss.tooling.spiderchart.style.AxisStyle;
import org.fortiss.tooling.spiderchart.style.ChartStyle;
import org.fortiss.tooling.spiderchart.style.DataSeriesStyle;
import org.fortiss.tooling.spiderchart.style.LegendStyle;

import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.chart.PieChart;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.control.SplitPane;
import javafx.scene.layout.BorderPane;

/**
 * FX Controller for the Metrics view.
 *
 * @author groh
 */

@SuppressWarnings("unchecked")
public class MetricsFXController extends CompositeFXControllerBase<SplitPane, Node>
		implements ISelectionListener {

	/** The {@link IProjectRootElement} in which the selected element is contained in. */
	private IProjectRootElement currentRootElement;

	/** The refresh {@link Button} to refresh the metrics. */
	@FXML
	private Button refreshButton;

	/** The central top {@link Label} to display short messages. */
	@FXML
	private Label topLabel;

	/** The bottom {@link BorderPane} for the spiderchart */
	@FXML
	private BorderPane borderPane;

	/** a pieChart */
	@FXML
	private PieChart pieChart;

	/** Time format when displaying updating time. */
	private final DateTimeFormatter timeFormat = DateTimeFormatter.ofPattern("HH:mm:ss");

	/** {@inheritDoc} */
	@Override
	public String getFXMLLocation() {
		return "MetricsViewFx.fxml";
	}

	/** {@inheritDoc} */
	@Override
	public void initialize() {
		// Not used

	}

	/** {@inheritDoc} */
	@Override
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {

		ModelQualityService.getInstance();
		var manager = ModelQualityService.metricDataManagerInstance;
		if(manager == null) {
			topLabel.setText(
					"No Metrics were available at " + timeFormat.format(LocalDateTime.now()));
			return;
		}

		EObject selected = checkAndPickFirst(selection, EObject.class);

		if(selected == null) {
			topLabel.setText("Nothing selected");
			return;
		}

		if(selected instanceof IProjectRootElement) {
			currentRootElement = (IProjectRootElement)selected;
		} else {
			currentRootElement = getFirstParentWithType(selected, IProjectRootElement.class);
		}

		if(currentRootElement == null) {
			// There was no root element for the selected element. (Usually should be the
			// FileProject.) This is not displayed here.
			topLabel.setText("No root found");
			return;
		}

		if(!(currentRootElement instanceof IHierarchicElement)) {
			topLabel.setText("Root is not an IHierarchicElement");
			return;
		}

		if(selected instanceof IHierarchicElement) {
			var root_provider = manager.getRootNodes().get(currentRootElement);

			if(root_provider == null) {
				topLabel.setText("No Metric found for this project, did you export metrics?");
				return;
			}

			var element = (IHierarchicElement)selected;
			var node = manager.getHierarchicLookupTable().get(element);
			if(node != null) {

				topLabel.setText("Text");

				if(!node.getChildren().isEmpty()) {

					SpiderChartViewer viewer = buildSpiderChart(node);
					borderPane.setCenter(viewer.getViewerPane());

					// Create PieChart and add data
					pieChart.getData().clear();
					for(var child : node.getChildren()) {

						PieChart.Data slice =
								new PieChart.Data(child.getName(), child.getTotalPortAmount());
						pieChart.getData().add(slice);
						pieChart.setLegendVisible(false);
						pieChart.setTitle("");
					}
				} else {
					borderPane.setCenter(null);
					pieChart.getData().clear();
				}

			} else {
				topLabel.setText("Element not in metrics");
			}
		} else {
			topLabel.setText("Selected is not an IHierarchicElement");
		}

	}

	/**
	 * Is triggered when the refresh button is pressed.
	 */
	public void onRefreshMetricsSubmit() {
		LocalDateTime currentTime = LocalDateTime.now();
		topLabel.setText("View was updated at " + timeFormat.format(currentTime));

	}

	/**
	 * @param provider
	 *            the provider for which the spider chart should be generated
	 * @return {@link SpiderChartViewer} with the matching metrics
	 */
	private static SpiderChartViewer buildSpiderChart(MetricTreeNode node) {
		SpiderChart spiderChart = new SpiderChart();
		spiderChart.setTitle("Amount of Ports");
		spiderChart.setLegendLabel("Legend");
		ChartStyle chartStyle = new ChartStyle(true, true, true);

		List<MetricTreeNode> children = node.getChildren();

		if(children.isEmpty()) {
			return null;
		}

		double maxval =
				children.stream().mapToDouble(p -> p.getTotalPortAmount()).max().getAsDouble();

		chartStyle.setUseIndividualAxisSegments(false);
		chartStyle.setTitleStyle(new FontStyle("Verdana", 14, BLUE.brighter()));

		DataSeries elementData = new DataSeries("Data");
		for(var child : node.getChildren()) {
			DoubleAxis testing = new DoubleAxis(child.getName(), 0.0, maxval);

			AxisStyle aStyle3Segs = new AxisStyle(SOLID_BLACK_1PT, BLACK_VERDANA_14PT, 3,
					BLACK_VERDANA_8PT, new DecimalFormat("#.##"));
			chartStyle.setAxisStyle(testing, aStyle3Segs);

			spiderChart.addAxis(testing);
			elementData.setPoint(testing, child.getTotalPortAmount());
		}
		spiderChart.addData(elementData);

		LineStyle olive1pt = new LineStyle(GREEN);
		FillStyle oliveFill = new FillStyle(GREEN, 0.25);
		DataSeriesStyle iphoneStyle = new DataSeriesStyle(olive1pt, oliveFill, true, true,
				BLACK_VERDANA_10PT, 7, new DecimalFormat("#.#"));
		chartStyle.setDataSeriesStyle(elementData, iphoneStyle);

		LegendStyle legendStyle = new LegendStyle(false, 5, BLACK_VERDANA_12PT);
		chartStyle.setLegendStyle(legendStyle);

		chartStyle.setAxisSegments(4);

		chartStyle.setBackgroundFillStyle(new FillStyle(LIGHTGRAY));
		chartStyle.setBackgroundLineStyle(new LineStyle(DARKGRAY));

		SpiderChartViewer viewer = new SpiderChartViewer(spiderChart, chartStyle);
		return viewer;
	}
}
